
package org.forester.rio;

public class RIOException extends Exception {

    /**
     *
     */
    private static final long serialVersionUID = 4691098852783522097L;

    public RIOException() {
        super();
    }

    public RIOException( final String message ) {
        super( message );
    }
}
