
package org.forester.util;

import java.awt.Color;

public final class TaxonomyColors {

    public final static Color DEUTEROSTOMIA_COLOR                    = new Color( 255, 0, 0 );
    public final static Color PROTOSTOMIA_COLOR                      = new Color( 204, 0, 0 );
    public final static Color CNIDARIA_COLOR                         = new Color( 204, 0, 142 );
    public final static Color PLACOZOA_COLOR                         = new Color( 204, 0, 132 );
    public final static Color CTENOPHORA_COLOR                       = new Color( 204, 0, 122 );
    public final static Color PORIFERA_COLOR                         = new Color( 204, 0, 112 );
    public final static Color CHOANOFLAGELLIDA                       = new Color( 135, 0, 255 );
    public final static Color ICHTHYOSPOREA_AND_FILASTEREA           = new Color( 125, 0, 225 );
    public final static Color DIKARYA_COLOR                          = new Color( 255, 155, 0 );
    public final static Color OTHER_FUNGI_COLOR                      = new Color( 227, 127, 0 );
    public final static Color NUCLEARIIDAE_AND_FONTICULA_GROUP_COLOR = new Color( 200, 100, 0 );
    public final static Color AMOEBOZOA_COLOR                        = new Color( 255, 0, 255 );
    public final static Color EMBRYOPHYTA_COLOR                      = new Color( 0, 255, 0 );
    public final static Color CHLOROPHYTA_COLOR                      = new Color( 0, 204, 0 );
    public final static Color RHODOPHYTA_COLOR                       = new Color( 0, 153, 76 );
    public final static Color HACROBIA_COLOR                         = new Color( 0, 90, 40 );
    public final static Color GLAUCOPHYTA_COLOR                      = new Color( 0, 90, 60 );
    public final static Color STRAMENOPILES_COLOR                    = new Color( 0, 0, 255 );
    public final static Color ALVEOLATA_COLOR                        = new Color( 0, 128, 255 );
    public final static Color RHIZARIA_COLOR                         = new Color( 0, 255, 255 );
    public static final Color APUSOZOA_COLOR                         = new Color( 204, 245, 245 );
    public final static Color EXCAVATA_COLOR                         = new Color( 204, 204, 0 );
    public final static Color ARCHAEA_COLOR                          = new Color( 160, 160, 160 );
    public final static Color BACTERIA_COLOR                         = new Color( 64, 64, 64 );
}
